@@ -44,11 +44,10 @@ from .utils import (
     create_dirichlet_bcs,
     create_config,
     load_config,
-    moreau_yosida_regularization,
 )
 
 
-__version__ = "1.4.4"
+__version__ = "1.5.0"
 
 __all__ = [
     "import_mesh",
@@ -63,7 +62,6 @@ __all__ = [
     "load_config",
     "create_bcs_list",
     "create_dirichlet_bcs",
-    "moreau_yosida_regularization",
     "verification",
     "ConstrainedOptimalControlProblem",
     "ConstrainedShapeOptimizationProblem",
