@@ -22,37 +22,46 @@ the "Hessian problems" occurring in the truncated Newton
 method.
 """
 
+from __future__ import annotations
+
 import abc
+from typing import TYPE_CHECKING, List, Optional
 
 import fenics
 import numpy as np
 from petsc4py import PETSc
 
-from .._exceptions import CashocsException, ConfigError, NotConvergedError
+from .._exceptions import CashocsException, NotConvergedError
 from .._loggers import debug
 from ..utils import _assemble_petsc_system, _setup_petsc_options, _solve_linear_problem
 
 
+if TYPE_CHECKING:
+    from .._forms import ControlFormHandler
+    from .control_gradient_problem import ControlGradientProblem
+
+
 class BaseHessianProblem(abc.ABC):
     """Base class for derived Hessian problems."""
 
-    def __init__(self, form_handler, gradient_problem):
-        """Initializes self.
-
+    def __init__(
+        self, form_handler: ControlFormHandler, gradient_problem: ControlGradientProblem
+    ) -> None:
+        """
         Parameters
         ----------
-        form_handler : cashocs._forms.ControlFormHandler
-                The FormHandler object for the optimization problem.
-        gradient_problem : cashocs._pde_problems.GradientProblem
-                The GradientProblem object (this is needed for the computation
-                of the Hessian).
+        form_handler : ControlFormHandler
+            The FormHandler object for the optimization problem.
+        gradient_problem : ControlGradientProblem
+            The ControlGradientProblem object (this is needed for the computation
+            of the Hessian).
         """
 
         self.form_handler = form_handler
         self.gradient_problem = gradient_problem
 
         self.config = self.form_handler.config
-        self.gradients = self.gradient_problem.gradients
+        self.gradient = self.gradient_problem.gradient
 
         self.inner_newton = self.config.get("AlgoTNM", "inner_newton", fallback="cr")
         self.max_it_inner_newton = self.config.getint(
@@ -133,7 +142,9 @@ class BaseHessianProblem(abc.ABC):
         for i, ksp in enumerate(self.ksps):
             ksp.setOperators(self.form_handler.riesz_projection_matrices[i])
 
-    def hessian_application(self, h, out):
+    def hessian_application(
+        self, h: List[fenics.Function], out: List[fenics.Function]
+    ) -> None:
         r"""Computes the application of the Hessian to some element
 
         This is needed in the truncated Newton method where we solve the system
@@ -144,10 +155,10 @@ class BaseHessianProblem(abc.ABC):
 
         Parameters
         ----------
-        h : list[dolfin.function.function.Function]
-                A function to which we want to apply the Hessian to.
-        out : list[dolfin.function.function.Function]
-                A list of functions into which the result is saved.
+        h : list[fenics.Function]
+            A function to which we want to apply the Hessian to.
+        out : list[fenics.Function]
+            A list of functions into which the result is saved.
 
         Returns
         -------
@@ -305,17 +316,19 @@ class BaseHessianProblem(abc.ABC):
 
         self.no_sensitivity_solves += 2
 
-    def newton_solve(self, idx_active=None):
+    def newton_solve(
+        self, idx_active: Optional[List[int]] = None
+    ) -> List[fenics.Function]:
         """Solves the problem with a truncated Newton method.
 
         Parameters
         ----------
-        idx_active : list[int]
+        idx_active : list[int] or None, optional
             List of active indices
 
         Returns
         -------
-         : list[dolfin.function.function.Function]
+        list[fenics.Function]
             The Newton increment
 
         """
@@ -330,82 +343,78 @@ class BaseHessianProblem(abc.ABC):
             self.cg(idx_active)
         elif self.inner_newton == "cr":
             self.cr(idx_active)
-        else:
-            raise ConfigError(
-                "AlgoTNM",
-                "inner_newton",
-                "Not a valid choice. Needs to be either 'cg' or 'cr'.",
-            )
 
         return self.delta_control
 
     @abc.abstractmethod
-    def cg(self, idx_active=None):
+    def cg(self, idx_active: Optional[List[int]] = None) -> None:
         """Solves the (truncated) Newton step with a CG method
 
         Parameters
         ----------
-        idx_active : list[int]
+        idx_active : list[int] or None, optional
             The list of active indices
 
         Returns
         -------
-         : list[dolfin.function.function.Function]
-            The Newton increment
-
+        None
         """
+
         pass
 
     @abc.abstractmethod
-    def cr(self, idx_active=None):
+    def cr(self, idx_active: Optional[List[int]] = None) -> None:
         """Solves the (truncated) Newton step with a CR method
 
         Parameters
         ----------
-        idx_active : list[int]
+        idx_active : list[int] or None, optional
             The list of active indices
 
         Returns
         -------
-         : list[dolfin.function.function.Function]
-            The Newton increment
-
+        None
         """
+
         pass
 
 
 class HessianProblem(BaseHessianProblem):
     """PDE Problem used to solve the (reduced) Hessian problem."""
 
-    def __init__(self, form_handler, gradient_problem):
+    def __init__(
+        self, form_handler: ControlFormHandler, gradient_problem: ControlGradientProblem
+    ) -> None:
         """Initializes self.
 
         Parameters
         ----------
-        form_handler : cashocs._forms.ControlFormHandler
-                The FormHandler object for the optimization problem.
-        gradient_problem : cashocs._pde_problems.GradientProblem
-                The GradientProblem object (this is needed for the computation
-                of the Hessian).
+        form_handler : ControlFormHandler
+            The FormHandler object for the optimization problem.
+        gradient_problem : ControlGradientProblem
+            The ControlGradientProblem object (this is needed for the computation
+            of the Hessian).
         """
 
         super().__init__(form_handler, gradient_problem)
 
-    def reduced_hessian_application(self, h, out):
+    def reduced_hessian_application(
+        self, h: List[fenics.Function], out: List[fenics.Function]
+    ) -> None:
         """Computes the application of the reduced Hessian on a direction.
 
         This is needed to solve the Newton step with iterative solvers.
 
         Parameters
         ----------
-        h : list[dolfin.function.function.Function]
+        h : list[fenics.Function]
             The direction, onto which the reduced Hessian is applied
-        out : list[dolfin.function.function.Function]
+        out : list[fenics.Function]
             The output of the application of the (linear) operator
 
         Returns
         -------
-
+        None
         """
 
         for j in range(self.control_dim):
@@ -425,19 +434,20 @@ class HessianProblem(BaseHessianProblem):
                 + self.inactive_part[j].vector().vec(),
             )
 
-    def newton_solve(self, idx_active=None):
+    def newton_solve(
+        self, idx_active: Optional[List[List[int]]] = None
+    ) -> List[fenics.Function]:
         """Solves the Newton step with an iterative method
 
         Parameters
         ----------
-        idx_active : list[list[int]]
+        idx_active : list[list[int]] or None, optional
             The list of active indices
 
         Returns
         -------
-         : list[dolfin.function.function.Function]
+        list[fenics.Function]
             The Newton increment
-
         """
 
         if idx_active is not None:
@@ -445,23 +455,23 @@ class HessianProblem(BaseHessianProblem):
 
         return super().newton_solve()
 
-    def cg(self, idx_active=None):
+    def cg(self, idx_active: Optional[List[int]] = None) -> None:
         """Solves the (truncated) Newton step with a CG method
 
         Parameters
         ----------
-        idx_active : list[int]
+        idx_active : list[int] or None, optional
             The list of active indices
 
         Returns
         -------
-         : list[dolfin.function.function.Function]
+        list[fenics.Function]
             The Newton increment
 
         """
 
         for j in range(self.control_dim):
-            self.residual[j].vector().vec().aypx(0.0, -self.gradients[j].vector().vec())
+            self.residual[j].vector().vec().aypx(0.0, -self.gradient[j].vector().vec())
             self.p[j].vector().vec().aypx(0.0, self.residual[j].vector().vec())
 
         self.rsold = self.form_handler.scalar_product(self.residual, self.residual)
@@ -503,23 +513,21 @@ class HessianProblem(BaseHessianProblem):
 
             self.rsold = self.rsnew
 
-    def cr(self, idx_active=None):
+    def cr(self, idx_active: Optional[List[int]] = None) -> None:
         """Solves the (truncated) Newton step with a CR method
 
         Parameters
         ----------
-        idx_active : list[int]
+        idx_active : list[int] or None, optional
             The list of active indices
 
         Returns
         -------
-         : list[dolfin.function.function.Function]
-            The Newton increment
-
+        None
         """
 
         for j in range(self.control_dim):
-            self.residual[j].vector().vec().aypx(0.0, -self.gradients[j].vector().vec())
+            self.residual[j].vector().vec().aypx(0.0, -self.gradient[j].vector().vec())
             self.p[j].vector().vec().aypx(0.0, self.residual[j].vector().vec())
 
         self.eps_0 = np.sqrt(
@@ -592,40 +600,49 @@ class HessianProblem(BaseHessianProblem):
 class UnconstrainedHessianProblem(BaseHessianProblem):
     """Hessian Problem without control constraints for the inner solver in PDAS."""
 
-    def __init__(self, form_handler, gradient_problem):
-        """Initializes self.
-
+    def __init__(
+        self, form_handler: ControlFormHandler, gradient_problem: ControlGradientProblem
+    ) -> None:
+        """
         Parameters
         ----------
-        form_handler : cashocs._forms.ControlFormHandler
-                The FormHandler object for the optimization problem.
-        gradient_problem : cashocs._pde_problems.GradientProblem
-                The GradientProblem object (this is needed for the computation
-                of the Hessian).
+        form_handler : ControlFormHandler
+            The FormHandler object for the optimization problem.
+        gradient_problem : ControlGradientProblem
+            The ControlGradientProblem object (this is needed for the computation
+            of the Hessian).
         """
 
         super().__init__(form_handler, gradient_problem)
 
         self.reduced_gradient = [
             fenics.Function(self.form_handler.control_spaces[j])
-            for j in range(len(self.gradients))
+            for j in range(len(self.gradient))
         ]
         self.temp = [fenics.Function(V) for V in self.form_handler.control_spaces]
 
-    def reduced_hessian_application(self, h, out, idx_active):
+    def reduced_hessian_application(
+        self,
+        h: List[fenics.Function],
+        out: List[fenics.Function],
+        idx_active: List[int],
+    ) -> None:
         """Computes the application of the reduced Hessian on a direction.
 
         This is needed to solve the Newton step with iterative solvers.
 
         Parameters
         ----------
-        h : list[dolfin.function.function.Function]
+        h : list[fenics.Function]
             The direction, onto which the reduced Hessian is applied
-        out : list[dolfin.function.function.Function]
+        out : list[fenics.Function]
             The output of the application of the (linear) operator
+        idx_active : list[int]
+            The list of active indices
 
         Returns
         -------
+        None
 
         """
 
@@ -638,17 +655,19 @@ class UnconstrainedHessianProblem(BaseHessianProblem):
         for j in range(self.control_dim):
             out[j].vector()[idx_active[j]] = 0.0
 
-    def newton_solve(self, idx_active=None):
+    def newton_solve(
+        self, idx_active: Optional[List[List[int]]] = None
+    ) -> List[fenics.Function]:
         """Solves the Newton step with an iterative method
 
         Parameters
         ----------
-        idx_active : list[list[int]]
+        idx_active : list[list[int]] or None, optional
             The list of active indices
 
         Returns
         -------
-         : list[dolfin.function.function.Function]
+        list[fenics.Function]
             The Newton increment
 
         """
@@ -662,25 +681,23 @@ class UnconstrainedHessianProblem(BaseHessianProblem):
 
         for j in range(self.control_dim):
             self.reduced_gradient[j].vector().vec().aypx(
-                0.0, self.gradients[j].vector().vec()
+                0.0, self.gradient[j].vector().vec()
             )
             self.reduced_gradient[j].vector()[idx_active[j]] = 0.0
 
         return super().newton_solve(idx_active)
 
-    def cg(self, idx_active=None):
+    def cg(self, idx_active: Optional[List[int]] = None) -> None:
         """Solves the (truncated) Newton step with a CG method
 
         Parameters
         ----------
-        idx_active : list[int]
+        idx_active : list[int] or None, optional
             The list of active indices
 
         Returns
         -------
-         : list[dolfin.function.function.Function]
-            The Newton increment
-
+        None
         """
 
         for j in range(self.control_dim):
@@ -719,19 +736,17 @@ class UnconstrainedHessianProblem(BaseHessianProblem):
 
             self.rsold = self.rsnew
 
-    def cr(self, idx_active=None):
+    def cr(self, idx_active: Optional[List[int]] = None) -> None:
         """Solves the (truncated) Newton step with a CR method
 
         Parameters
         ----------
-        idx_active : list[int]
+        idx_active : list[int] or None, optional
             The list of active indices
 
         Returns
         -------
-         : list[dolfin.function.function.Function]
-            The Newton increment
-
+        None
         """
 
         for j in range(self.control_dim):
