@@ -21,31 +21,46 @@ This class uses the linear elasticity equations to project the
 shape derivative to the shape gradient with a Riesz projection.
 """
 
+from __future__ import annotations
+
+import configparser
+from typing import TYPE_CHECKING, List
+
 import fenics
 import numpy as np
+import ufl
 from petsc4py import PETSc
 
-from .._exceptions import ConfigError
 from .._interfaces.pde_problem import PDEProblem
 from ..nonlinear_solvers import newton_solve
 from ..utils import _setup_petsc_options, _solve_linear_problem
 
 
+if TYPE_CHECKING:
+    from .state_problem import StateProblem
+    from .adjoint_problem import AdjointProblem
+    from .._forms import ShapeFormHandler
+
 
 class ShapeGradientProblem(PDEProblem):
     """Riesz problem for the computation of the shape gradient."""
 
-    def __init__(self, form_handler, state_problem, adjoint_problem):
+    def __init__(
+        self,
+        form_handler: ShapeFormHandler,
+        state_problem: StateProblem,
+        adjoint_problem: AdjointProblem,
+    ) -> None:
         """Initialize the ShapeGradientProblem.
 
         Parameters
         ----------
-        form_handler : cashocs._forms.ShapeFormHandler
-                The ShapeFormHandler object corresponding to the shape optimization problem.
-        state_problem : cashocs._pde_problems.StateProblem
-                The corresponding state problem.
-        adjoint_problem : cashocs._pde_problems.AdjointProblem
-                The corresponding adjoint problem.
+        form_handler : ShapeFormHandler
+            The ShapeFormHandler object corresponding to the shape optimization problem.
+        state_problem : StateProblem
+            The corresponding state problem.
+        adjoint_problem : AdjointProblem
+            The corresponding adjoint problem.
         """
 
         super().__init__(form_handler)
@@ -83,12 +98,7 @@ class ShapeGradientProblem(PDEProblem):
                 ["ksp_atol", 1e-50],
                 ["ksp_max_it", 250],
             ]
-        else:
-            raise ConfigError(
-                "OptimizationRoutine",
-                "gradient_method",
-                "gradient_method has to be either 'direct' or 'iterative'",
-            )
+
         _setup_petsc_options([self.ksp], [self.ksp_options])
 
         if (
@@ -103,13 +113,13 @@ class ShapeGradientProblem(PDEProblem):
                 self.config,
             )
 
-    def solve(self):
+    def solve(self) -> fenics.Function:
         """Solves the Riesz projection problem to obtain the shape gradient of the cost functional.
 
         Returns
         -------
-        gradient : dolfin.function.function.Function
-                The function representing the shape gradient of the (reduced) cost functional.
+        fenics.Function
+            The function representing the shape gradient of the (reduced) cost functional.
         """
 
         self.state_problem.solve()
@@ -117,7 +127,7 @@ class ShapeGradientProblem(PDEProblem):
 
         if not self.has_solution:
 
-            self.form_handler.regularization.update_geometric_quantities()
+            self.form_handler.shape_regularization.update_geometric_quantities()
 
             if (
                 self.config.getboolean(
@@ -136,14 +146,18 @@ class ShapeGradientProblem(PDEProblem):
                 self.form_handler.assembler.assemble(
                     self.form_handler.fe_shape_derivative_vector
                 )
+                if self.form_handler.use_fixed_dimensions:
+                    self.form_handler.fe_shape_derivative_vector[
+                        self.form_handler.fixed_indices
+                    ] = 0.0
                 _solve_linear_problem(
                     self.ksp,
                     self.form_handler.scalar_product_matrix,
                     self.form_handler.fe_shape_derivative_vector.vec(),
-                    self.gradient.vector().vec(),
+                    self.gradient[0].vector().vec(),
                     self.ksp_options,
                 )
-                self.gradient.vector().apply("")
+                self.gradient[0].vector().apply("")
 
                 self.has_solution = True
 
@@ -157,21 +171,44 @@ class ShapeGradientProblem(PDEProblem):
 
 
 class _PLaplacProjector:
+    """A class for computing the gradient deformation with a p-Laplace projection"""
+
     def __init__(
-        self, shape_gradient_problem, gradient, shape_derivative, bcs_shape, config
-    ):
+        self,
+        gradient_problem: ShapeGradientProblem,
+        gradient: List[fenics.Function],
+        shape_derivative: ufl.Form,
+        bcs_shape: List[fenics.DirichletBC],
+        config: configparser.ConfigParser,
+    ) -> None:
+        """
+
+        Parameters
+        ----------
+        gradient_problem : ShapeGradientProblem
+            The shape gradient problem
+        gradient : list[fenics.Function]
+            The fenics Function representing the gradient deformation
+        shape_derivative : ufl.Form
+            The ufl Form of the shape derivative
+        bcs_shape : list[fenics.DirichletBC]
+            The boundary conditions for computing the gradient deformation
+        config : configparser.ConfigParser
+            The config for the optimization problem
+        """
+
         self.p_target = config.getint("ShapeGradient", "p_laplacian_power", fallback=2)
         delta = config.getfloat("ShapeGradient", "damping_factor", fallback=0.0)
         eps = config.getfloat(
             "ShapeGradient", "p_laplacian_stabilization", fallback=0.0
         )
         self.p_list = np.arange(2, self.p_target + 1, 1)
-        self.solution = gradient
+        self.solution = gradient[0]
         self.shape_derivative = shape_derivative
-        self.test_vector_field = shape_gradient_problem.form_handler.test_vector_field
+        self.test_vector_field = gradient_problem.form_handler.test_vector_field
         self.bcs_shape = bcs_shape
-        dx = shape_gradient_problem.form_handler.dx
-        self.mu_lame = shape_gradient_problem.form_handler.mu_lame
+        dx = gradient_problem.form_handler.dx
+        self.mu_lame = gradient_problem.form_handler.mu_lame
 
         self.A_tensor = fenics.PETScMatrix()
         self.b_tensor = fenics.PETScVector()
@@ -215,24 +252,17 @@ class _PLaplacProjector:
                     ["ksp_atol", 1e-50],
                     ["ksp_max_it", 100],
                 ]
-            else:
-                raise ConfigError(
-                    "OptimizationRoutine",
-                    "gradient_method",
-                    "gradient_method has to be either 'direct' or 'iterative'",
-                )
 
-    def solve(self):
+    def solve(self) -> None:
         """Solves the p-Laplace problem for computing the shape gradient
 
         Returns
         -------
-
+        None
         """
 
         self.solution.vector().vec().set(0.0)
         for F in self.F_list:
-
             newton_solve(
                 F,
                 self.solution,
