@@ -19,15 +19,14 @@
 
 """
 
-from .regularization import Regularization
 from .shape_cost_functional import ReducedShapeCostFunctional
 from .shape_line_search import ArmijoLineSearch
 from .shape_optimization_algorithm import ShapeOptimizationAlgorithm
-
+from .shape_optimization_problem import ShapeOptimizationProblem
 
 __all__ = [
-    "Regularization",
     "ReducedShapeCostFunctional",
     "ArmijoLineSearch",
     "ShapeOptimizationAlgorithm",
+    "ShapeOptimizationProblem",
 ]
