@@ -22,64 +22,71 @@ which are the :math:`L^2` distances between current volume, surface,
 and barycenter, and desired ones.
 """
 
+from __future__ import annotations
+
 import json
+from typing import TYPE_CHECKING
 
 import fenics
+import ufl
 from fenics import Constant, div, inner
 
-from .._exceptions import ConfigError
 from .._loggers import info
-from ..utils import _solve_linear_problem, _check_for_config_list
+from ..utils import _solve_linear_problem
 
 
-def t_grad(u, n):
+if TYPE_CHECKING:
+    from .._forms import ShapeFormHandler
+
+
+def t_grad(u: fenics.Function, n: fenics.FacetNormal) -> ufl.core.expr.Expr:
     """Computes the tangential gradient of u
 
     Parameters
     ----------
-    u : dolfin.function.function.Function
-            the argument
-    n : ufl.geometry.FacetNormal
-            the unit outer normal vector
+    u : fenics.Function
+        the argument
+    n : fenics.FacetNormal
+        the unit outer normal vector
 
     Returns
     -------
     ufl.core.expr.Expr
-            the tangential gradient of u
+        the tangential gradient of u
     """
 
     return fenics.grad(u) - fenics.outer(fenics.grad(u) * n, n)
 
 
-def t_div(u, n):
-    """Computes the tangential divergence
+def t_div(u: fenics.Function, n: fenics.FacetNormal) -> ufl.core.expr.Expr:
+    """Computes the tangential divergence of u
 
     Parameters
     ----------
-    u : dolfin.function.function.Function
-            the argument
-    n : ufl.geometry.FacetNormal
-            the outer unit normal vector
+    u : fenics.Function
+        the argument
+    n : fenics.FacetNormal
+        the outer unit normal vector
 
     Returns
     -------
     ufl.core.expr.Expr
-            the tangential divergence of u
+        the tangential divergence of u
     """
 
     return fenics.div(u) - fenics.inner(fenics.grad(u) * n, n)
 
 
-class Regularization:
+class OldRegularization:
     """Regularization terms for shape optimization problems"""
 
-    def __init__(self, form_handler):
+    def __init__(self, form_handler: ShapeFormHandler) -> None:
         """Initializes the regularization
 
         Parameters
         ----------
-        form_handler : cashocs._forms.ShapeFormHandler
-                the corresponding shape form handler object
+        form_handler : ShapeFormHandler
+            the corresponding shape form handler object
         """
 
         self.test_vector_field = form_handler.test_vector_field
@@ -110,8 +117,12 @@ class Regularization:
             )
             self.x_end = self.config.getfloat("Regularization", "x_end", fallback=1.0)
             if not self.x_end >= self.x_start:
-                raise ConfigError(
-                    "Regularization", "x_end", "x_end must not be smaller than x_start."
+                raise IncompatibleConfigurationError(
+                    "x_start",
+                    "Regularization",
+                    "x_end",
+                    "Regularization",
+                    "Reason: x_end must not be smaller than x_start.",
                 )
             self.delta_x = self.x_end - self.x_start
 
@@ -120,8 +131,12 @@ class Regularization:
             )
             self.y_end = self.config.getfloat("Regularization", "y_end", fallback=1.0)
             if not self.y_end >= self.y_start:
-                raise ConfigError(
-                    "Regularization", "y_end", "y_end must not be smaller than y_start."
+                raise IncompatibleConfigurationError(
+                    "y_start",
+                    "Regularization",
+                    "y_end",
+                    "Regularization",
+                    "Reason: y_end must not be smaller than y_start.",
                 )
             self.delta_y = self.y_end - self.y_start
 
@@ -130,8 +145,12 @@ class Regularization:
             )
             self.z_end = self.config.getfloat("Regularization", "z_end", fallback=1.0)
             if not self.z_end >= self.z_start:
-                raise ConfigError(
-                    "Regularization", "z_end", "z_end must not be smaller than z_start."
+                raise IncompatibleConfigurationError(
+                    "z_start",
+                    "Regularization",
+                    "z_end",
+                    "Regularization",
+                    "Reason: z_end must not be smaller than z_start.",
                 )
             self.delta_z = self.z_end - self.z_start
             if self.geometric_dimension == 2:
@@ -194,17 +213,7 @@ class Regularization:
         target_bar = self.config.get(
             "Regularization", "target_barycenter", fallback="[0,0,0]"
         )
-        if not _check_for_config_list(target_bar):
-            raise ConfigError(
-                "Regularization",
-                "target_barycenter",
-                "The target_barycenter parameter has to be a list of numbers",
-            )
         self.target_barycenter_list = json.loads(target_bar)
-        if not isinstance(self.target_barycenter_list, list):
-            raise ConfigError(
-                "Regularization", "target_barycenter", "This has to be a list."
-            )
 
         if self.geometric_dimension == 2 and len(self.target_barycenter_list) == 2:
             self.target_barycenter_list.append(0.0)
@@ -257,18 +266,6 @@ class Regularization:
                 else:
                     self.target_barycenter_list[2] = 0.0
 
-        if not (
-            self.mu_volume >= 0.0
-            and self.mu_surface >= 0.0
-            and self.mu_curvature >= 0
-            and self.mu_barycenter >= 0.0
-        ):
-            raise ConfigError(
-                "Regularization",
-                "mu_volume, mu_surface, or mu_barycenter",
-                "All regularization constants have to be nonnegative.",
-            )
-
         if (
             self.mu_volume > 0.0
             or self.mu_surface > 0.0
@@ -287,7 +284,7 @@ class Regularization:
         self.current_barycenter_y = fenics.Expression("val", degree=0, val=0.0)
         self.current_barycenter_z = fenics.Expression("val", degree=0, val=0.0)
 
-    def update_geometric_quantities(self):
+    def update_geometric_quantities(self) -> None:
         """Updates the geometric quantities
 
         Updates the volume, surface area, and barycenters (after the
@@ -352,7 +349,7 @@ class Regularization:
 
         self.compute_curvature()
 
-    def compute_curvature(self):
+    def compute_curvature(self) -> None:
         """Computes the mean curvature vector of the geometry.
 
         Returns
@@ -377,13 +374,13 @@ class Regularization:
         else:
             pass
 
-    def compute_objective(self):
+    def compute_objective(self) -> float:
         """Computes the part of the objective value that comes from the regularization
 
         Returns
         -------
         float
-                Part of the objective value coming from the regularization
+            Part of the objective value coming from the regularization
 
         """
 
@@ -478,13 +475,13 @@ class Regularization:
         else:
             return 0.0
 
-    def compute_shape_derivative(self):
+    def compute_shape_derivative(self) -> ufl.Form:
         """Computes the part of the shape derivative that comes from the regularization
 
         Returns
         -------
-        ufl.form.Form
-                The weak form of the shape derivative coming from the regularization
+        ufl.Form
+            The weak form of the shape derivative coming from the regularization
 
         """
 
@@ -618,6 +615,12 @@ class Regularization:
             return inner(fenics.Constant([0] * dim), V) * self.dx
 
     def _scale_weights(self):
+        """Scales the terms of the regularization by the weights given in the config file
+
+        Returns
+        -------
+        None
+        """
 
         if self.use_relative_scaling and self.has_regularization:
 
